typedef Template<T> = String Function(T input);

typedef AttributeExtractor<T> = Map<String, dynamic> Function(T input);

abstract class Widget<T> {
  String call([T input]);
}

class TemplateWidget<T> implements Widget<T> {
  final Template<T> template;

  TemplateWidget(this.template);

  @override
  String call([T input]) => template(input);
}

class ElementWidget<T> extends TemplateWidget<T> {
  final String element;
  final Map<String, dynamic> _attributes;
  final AttributeExtractor<T> dynamicAttributes;
  final bool collapse;
  final bool newlineAfterStart;
  final bool newlineAfterEnd;

  ElementWidget(
    this.element, {
    String id,
    Map<String, String> style,
    List<String> classes,
    Template<T> template = _defaultTemplate,
    Map<String, dynamic> attributes = const {},
    this.collapse = false,
    this.newlineAfterStart = false,
    this.newlineAfterEnd = true,
    this.dynamicAttributes = _defaultAttributeExtractor,
  })  : _attributes = {},
        super(template) {
    if (id != null) {
      _attributes['id'] = id;
    }
    if (classes?.isNotEmpty ?? false) {
      _attributes['class'] = _formatClasses(classes);
    }
    if (style?.isNotEmpty ?? false) {
      _attributes['style'] = _formatStyle(style);
    }
    if (attributes != null) {
      _attributes.addAll(attributes);
    }
  }

  @override
  String call([T input]) {
    var content = template(input);
    if (content.isNotEmpty || !collapse) {
      return '<$element${_renderAttributes(input)}>'
          '${newlineAfterStart && content.isNotEmpty ? '\n' : ''}'
          '$content'
          '</$element>'
          '${newlineAfterEnd ? '\n' : ''}';
    }
    return '<$element${_renderAttributes(input)}>${newlineAfterEnd ? '\n' : ''}';
  }

  String _renderAttributes(T input) {
    _attributes.addAll(dynamicAttributes(input));
    var output = _transformAttributes();
    return output.isEmpty ? '' : ' $output';
  }

  String _transformAttributes() {
    return _attributes.entries
        .where((entry) => entry.value != null)
        .map((entry) => _renderAttribute(entry.key, entry.value))
        .where((output) => output != '')
        .join(' ');
  }

  String _renderAttribute(String name, dynamic value) {
    if (value is bool) {
      return value ? '$name' : '';
    }
    return '${name}="${value}"';
  }

  static String _defaultTemplate(dynamic input) {
    var output = (input is Iterable)
        ? input.where((item) => item != null).map((item) => item.endsWith('\n') ? item : '$item\n').join()
        : input;
    return '${output ?? ''}';
  }

  static Map<String, dynamic> _defaultAttributeExtractor(dynamic input) => {};

  String _formatClasses(List<String> classes) => classes?.join(' ');

  String _formatStyle(Map<String, String> style) =>
      style?.entries?.map((entry) => '${entry.key}: ${entry.value}')?.join('; ');
}
